// FHE.cpp : Defines the entry point for the console application.
// 
/*This code is by Yongge Wang, November 2015.
the code is for FHE scheme
this code requries NTL library which could be obtained
at http://shoup.net/ntl/ 
*/

#include <tchar.h>
#include <string>
#include <NTL/mat_ZZ_p.h>
#include "FHETransformCipher.h"

using namespace std;
using namespace NTL;

void FHEOperationsTestSuite();
ZZ testFHEAddition(const ZZ& a, const ZZ& b);
ZZ testFHEMultiplication(const ZZ& a, const ZZ& b);

void verifyCipher();

void printTestResults(const string& testname, const ZZ& expected, const ZZ& actual);
void readKey();

int _tmain(int argc, _TCHAR* argv[])
//int _tmain(int argc, char* argv[])
{

	FHEOperationsTestSuite();
	//verifyCipher();

	readKey();
	return 0;
}


void verifyCipher(){
	FHETransformCipher fhe;
	FHEKey key = FHETransformCipher::generateKey();
	key.printKey();
	FHETransformCipher::verifyKey(key);
}

void FHEOperationsTestSuite(){
	ZZ a = to_ZZ(RandomBnd(10000));
	ZZ b = to_ZZ(RandomBnd(10000));

	ZZ sum, product;
	product = to_ZZ(1);


	printTestResults(string("FHE Addition(A+B)"), a + b, testFHEAddition(a, b));
	printTestResults(string("FHE Multiplication(A*B)"), a * b, testFHEMultiplication(a, b));

}

ZZ testFHEAddition(const ZZ& a, const ZZ& b){
	FHETransformCipher fhe;
	FHEKey key = FHETransformCipher::generateKey();

	mat_ZZ_p Cm0, Cm1;
	ZZ result;

	fhe.encrypt(key, a, Cm0);
	fhe.encrypt(key, b, Cm1);
	Cm0 += Cm1;
	fhe.decrypt(key, Cm0, result);

	return result;
}

ZZ testFHEMultiplication(const ZZ& a, const ZZ& b){
	FHETransformCipher fhe;
	FHEKey key = FHETransformCipher::generateKey();

	mat_ZZ_p Cm0, Cm1;
	ZZ result;

	fhe.encrypt(key, a, Cm0);
	fhe.encrypt(key, b, Cm1);
	Cm0 *= Cm1;
	fhe.decrypt(key, Cm0, result);

	return result;
}



void printTestResults(const string& testname, const ZZ& expected, const ZZ& actual){
	cout << testname << "\n";
	cout << "Expected: " << expected << "\n";
	cout << "Actual:   " << actual << "\n";
	cout << "\n";
}

void readKey(){
	int k_in;
	cin >> k_in;
}

