#include "FHETransformCipher.h"
#include <NTL/mat_ZZ_p.h>
#include <NTL/ZZ.h>
#include <NTL/tools.h>
#include <iostream>
#include <fstream>


FHETransformCipher::FHETransformCipher(){}

FHETransformCipher::~FHETransformCipher(){}


int FHETransformCipher::encrypt(const FHEKey& key, const ZZ& message, mat_ZZ_p& Cm){
	vec_ZZ_p m_vec;
	m_vec.SetLength(MAX_VEC_LENGTH);

	if (message >= key.n){
		return 1;
	}

	ZZ_p::init(key.n);
	ZZ_p a = NTL::to_ZZ_p(NTL::RandomBnd(key.n - 1));
	vec_ZZ_p ar = a*key.r;

	m_vec.put(0, to_ZZ_p(message));
	m_vec += ar;

	mat_ZZ_p Am = vectorToOctonion(m_vec);

	Cm = ((key.K*Am)*NTL::inv(key.K));

	return 0;
}

int FHETransformCipher::decrypt(const FHEKey& key, const mat_ZZ_p& Cm, ZZ& message){
	ZZ_p::init(key.n);
	mat_ZZ_p Am;
	vec_ZZ_p m_vec, a_vec;

	m_vec.SetLength(MAX_VEC_LENGTH);
	a_vec.SetLength(MAX_VEC_LENGTH);

	Am = NTL::inv(key.K)*Cm*key.K;
	m_vec = octonionToVector(Am);
	for (int i = 0; i < MAX_VEC_LENGTH; i++){
		a_vec.put(i, m_vec.get(i) / key.r.get(i));
	}

	m_vec -= (a_vec.get(1)*key.r);
	conv(message, m_vec.get(0));

	return 0;
}

FHEKey FHETransformCipher::generateKey(){
	ZZ p, q;
	// vec_ZZ_p r = getStaticIsotropicVector(p, q);
	vec_ZZ_p r = getIsotropicVector(p, q);
	ZZ n = p*q;
	mat_ZZ_p m = generateInvertableMatrix(n);

	return FHEKey(m, r, n);
}

bool FHETransformCipher::verifyKey(const FHEKey& key){
	cout << "Isotropic Vector of length " << key.r.length() << "\n";

	if (verifyIsotropicVector(key.r, key.n)){
		cout << "Key.r is a valid Isotropic Vector\n";
	}

	mat_ZZ_p i;
	ZZ_p::init(key.n);
	i.SetDims(MAX_VEC_LENGTH, MAX_VEC_LENGTH);
	inv(i, key.K);
	cout << "Key.K is an inverable matrix in mod n\n";
	return true;
}


/*Private Class Methods*/
ZZ_p FHETransformCipher::negate(const ZZ_p& a){
	ZZ_p x;
	NTL::negate(x, a);
	return x;
}

mat_ZZ_p FHETransformCipher::vectorToOctonion(const vec_ZZ_p& m){
	mat_ZZ_p oct;
	oct.SetDims(MAX_VEC_LENGTH, MAX_VEC_LENGTH);

	oct.put(0, 0, m.get(0));
	oct.put(0, 1, m.get(1));
	oct.put(0, 2, m.get(2));
	oct.put(0, 3, m.get(3));
	oct.put(0, 4, m.get(4));
	oct.put(0, 5, m.get(5));
	oct.put(0, 6, m.get(6));
	oct.put(0, 7, m.get(7));

	oct.put(1, 0, negate(m.get(1)));
	oct.put(1, 1, m.get(0));
	oct.put(1, 2, m.get(4));
	oct.put(1, 3, m.get(7));
	oct.put(1, 4, negate(m.get(2)));
	oct.put(1, 5, m.get(6));
	oct.put(1, 6, negate(m.get(5)));
	oct.put(1, 7, negate(m.get(3)));

	oct.put(2, 0, negate(m.get(2)));
	oct.put(2, 1, negate(m.get(4)));
	oct.put(2, 2, m.get(0));
	oct.put(2, 3, m.get(5));
	oct.put(2, 4, m.get(1));
	oct.put(2, 5, negate(m.get(3)));
	oct.put(2, 6, m.get(7));
	oct.put(2, 7, negate(m.get(6)));

	oct.put(3, 0, negate(m.get(3)));
	oct.put(3, 1, negate(m.get(7)));
	oct.put(3, 2, negate(m.get(5)));
	oct.put(3, 3, m.get(0));
	oct.put(3, 4, m.get(6));
	oct.put(3, 5, m.get(2));
	oct.put(3, 6, negate(m.get(4)));
	oct.put(3, 7, m.get(1));

	oct.put(4, 0, negate(m.get(4)));
	oct.put(4, 1, m.get(2));
	oct.put(4, 2, negate(m.get(1)));
	oct.put(4, 3, negate(m.get(6)));
	oct.put(4, 4, m.get(0));
	oct.put(4, 5, m.get(7));
	oct.put(4, 6, m.get(3));
	oct.put(4, 7, negate(m.get(5)));

	oct.put(5, 0, negate(m.get(5)));
	oct.put(5, 1, negate(m.get(6)));
	oct.put(5, 2, m.get(3));
	oct.put(5, 3, negate(m.get(2)));
	oct.put(5, 4, negate(m.get(7)));
	oct.put(5, 5, m.get(0));
	oct.put(5, 6, m.get(1));
	oct.put(5, 7, m.get(4));

	oct.put(6, 0, negate(m.get(6)));
	oct.put(6, 1, m.get(5));
	oct.put(6, 2, negate(m.get(7)));
	oct.put(6, 3, m.get(4));
	oct.put(6, 4, negate(m.get(3)));
	oct.put(6, 5, negate(m.get(1)));
	oct.put(6, 6, m.get(0));
	oct.put(6, 7, m.get(2));

	oct.put(7, 0, negate(m.get(7)));
	oct.put(7, 1, m.get(3));
	oct.put(7, 2, m.get(6));
	oct.put(7, 3, negate(m.get(1)));
	oct.put(7, 4, m.get(5));
	oct.put(7, 5, negate(m.get(4)));
	oct.put(7, 6, negate(m.get(2)));
	oct.put(7, 7, m.get(0));

	return oct;
}

vec_ZZ_p FHETransformCipher::octonionToVector(const mat_ZZ_p& oct){
	vec_ZZ_p m;
	m.SetLength(MAX_VEC_LENGTH);
	for (int i = 0; i < MAX_VEC_LENGTH; i++){
		m.put(i, oct.get(0, i));
	}
	return m;
}

bool FHETransformCipher::verifyIsotropicVector(const vec_ZZ_p& r, const ZZ& n){
	ZZ TWO = to_ZZ(2);
	ZZ_p ZERO = to_ZZ_p(0);
	ZZ_p sum = to_ZZ_p(0);
	ZZ_p::init(n);

	for (int i = 0; i < MAX_VEC_LENGTH; i++){
		sum += power(r.get(i), TWO);
	}

	return (sum == ZERO) ? true : false;
}

vec_ZZ_p FHETransformCipher::getStaticIsotropicVector(ZZ& p, ZZ&q){
	vec_ZZ_p r;
	

	p = to_ZZ(40739);
	q = to_ZZ(65171);
	ZZ n = p*q;
	ZZ_p::init(n);
	r.SetLength(MAX_VEC_LENGTH);

	ZZ hack = to_ZZ(1158128019);
	hack = hack * 2;

	r.put(0, to_ZZ_p(125678544));
	r.put(1, to_ZZ_p(1637087964));
	r.put(2, to_ZZ_p(207527336));
	r.put(3, to_ZZ_p(1661101135));
	r.put(4, to_ZZ_p(1011490717));
	r.put(5, to_ZZ_p(61230684));
	r.put(6, to_ZZ_p(453449215));
	r.put(7, to_ZZ_p(hack));

	return r;
}


vec_ZZ_p FHETransformCipher::getIsotropicVector(ZZ& p, ZZ&q){
	vec_ZZ_p r;

	bool p4 = 0;
	int remainder = 0;
	while (!p4) {
		GenPrime(p, 512, 80);
		remainder = p % 4;
		if (remainder == 3) {
			p4 = 1;
		}
	}
	p4 = 0;
	remainder = 0;
	while (!p4) {
		GenPrime(q, 512, 80);
		remainder = q % 4;
		if (remainder == 3) {
			p4 = 1;
		}
	}
	ZZ n = p*q;
	ZZ_p::init(n);
	r.SetLength(MAX_VEC_LENGTH);
	ZZ_p rSq = ZZ_p();
	ZZ_p temprSq = ZZ_p();
	random(r[0]);
	rSq = rSq + sqr(r[0]);
	random(r[1]);
	rSq = rSq + sqr(r[1]);
	random(r[2]);
	rSq = rSq + sqr(r[2]);
	random(r[3]);
	rSq = rSq + sqr(r[3]);
	random(r[4]);
	rSq = rSq + sqr(r[4]);
	random(r[5]);
	rSq = rSq + sqr(r[5]);

	bool isotropic = 0;
	while (!isotropic) {
		temprSq = rSq;
		random(r[6]);
		temprSq = temprSq + sqr(r[6]);
		temprSq = negate(temprSq);

		ZZ rSqZZ, mp, mq;
		rSqZZ = temprSq._ZZ_p__rep;
		/* the following process not work, not know why
		mp = PowerMod(rSqZZ, (p + 1) / 4, p);
		mq = PowerMod(rSqZZ, (q + 1) / 4, q);
		*/

		//try my own PowerMod
		ZZ pOrder, qOrder;
		pOrder = (p + 1) / 4;
		qOrder = (q + 1) / 4;
		long k = NumBits(pOrder);
		mp = 1;
		for (long i = k - 1; i >= 0; i--) {
			mp = (mp*mp) % p;
			if (bit(pOrder, i) == 1) mp = (mp*rSqZZ) % p;
		}
		mq = 1;
		k = NumBits(qOrder);
		mq = 1;
		for (long i = k - 1; i >= 0; i--) {
			mq = (mq*mq) % q;
			if (bit(qOrder, i) == 1) mq = (mq*rSqZZ) % q;
		}
		//end my own PowerMod

		ZZ d, yp, yq, r7;

		XGCD(d, yp, yq, p, q);
		AddMod(r7, yp*p*mq, yq*q*mp, n);
		r.put(7, to_ZZ_p(r7));

		ZZ_p test;
		test = 0;
		for (int i = 0; i < 8; i++) {
			test = test + sqr(r[i]);
			cout << "r[" << i << "] is " << r[i] << endl;
		}
		if (test == 0) {
			isotropic = 1;
		}
	}
	cout << "n = " << n << endl;
	cout << "p= " << p << endl;
	cout << "q= " << q << endl;
	
	return r;
}

mat_ZZ_p FHETransformCipher::generateInvertableMatrix(const ZZ& n){
	mat_ZZ_p m, i;
	
	ZZ_p::init(n);
	i.SetDims(MAX_VEC_LENGTH, MAX_VEC_LENGTH);
	m.SetDims(MAX_VEC_LENGTH, MAX_VEC_LENGTH);
	for (int j = 0; j < MAX_VEC_LENGTH; j++){
		for (int k = 0; k < MAX_VEC_LENGTH; k++){
			m.put(j, k, to_ZZ_p(RandomBnd(n - 1)));
		}
	}
	inv(i, m);
	return m;
}



