#pragma once

#include <NTL/ZZ_p.h>
#include <NTL/vec_ZZ_p.h>
#include <NTL/mat_ZZ_p.h>

using namespace std;
using namespace NTL;

const long MAX_VEC_LENGTH = 8;
const long MAX_BIT_LENGTH = 32;

struct FHEKey{
	mat_ZZ_p K;
	vec_ZZ_p r;
	ZZ n;

	FHEKey(){}
	FHEKey(mat_ZZ_p& K, vec_ZZ_p& r, ZZ& n) : K(K), r(r), n(n){}

	void FHEKey::printKey(){
		cout << "Modulo n = " << n << "\n";
		cout << "Isotropic Vector: " << r << "\n";
		cout << "Invertable Matrix:\n" << K << "\n"; 
	}
};

class FHETransformCipher
{
public:
	FHETransformCipher();
	~FHETransformCipher();
	int encrypt(const FHEKey& key, const ZZ& message, mat_ZZ_p& Cm);
	int decrypt(const FHEKey& key, const mat_ZZ_p& Cm, ZZ& message);
	static FHEKey generateKey();
	static bool verifyKey(const FHEKey& key);

private:
	static ZZ_p negate(const ZZ_p& a);
	static mat_ZZ_p vectorToOctonion(const vec_ZZ_p& m);
	static vec_ZZ_p octonionToVector(const mat_ZZ_p& oct);
	static bool verifyIsotropicVector(const vec_ZZ_p& r, const ZZ& n);
	static vec_ZZ_p getStaticIsotropicVector(ZZ& p, ZZ& q);
	static vec_ZZ_p getIsotropicVector(ZZ& p, ZZ& q);
	static mat_ZZ_p generateInvertableMatrix(const ZZ& n);
};

